/*
 * TCSS 305Winter 2014
 * Assignment 5b-PowerPaint
 */

package drawingpanelandtools;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.event.MouseInputAdapter;

/**
 * This class constructs the JFrame and starts the construction of the various panels.
 * 
 * @author Brian Dennis
 * @version 20 February 2014
 */
@SuppressWarnings("serial")
public class DrawingPanel extends JPanel {

    /**
     * The panel width.
     */
    private static final int WIDTH = 500;

    /**
     * The panel height.
     */
    private static final int HEIGHT = 300;

    /**
     * The space between the grid segments.
     */
    private static final int GRID_WIDTH = 10;
    
    /**
     * The background color of the panel.
     */
    private static final Color CANVAS_COLOR = Color.WHITE;
    
    /** A string used to represent the command for the Undo property change. */
    private static final String UNDO_TOGGLE = "ToggleUndo";
    
    /** A string used to represent the command for the Redo property change. */
    private static final String REDO_TOGGLE = "ToggleRedo";

    /** An ArrayList that holds all the shapes that need to be redrawn on the board. */
    private final List<PaintShape> myShapes;
    
    /** This is an ArrayDeque that functions as a stack to hold shapes for undoing changes. */
    private final ArrayDeque<PaintShape> myUndoStack;
    
    /** This is an ArrayDeque that functions as a stack to hold shapes for redoing changes. */
    private final ArrayDeque<PaintShape> myRedoStack;
    
    /** the color of the current selection. */
    private Color myBrushColor;
    
    /** the brush width of the current selection. */
    private int myBrushWidth;
    
    /** the current tool that is being used. */
    private Tool myCurrentTool;
    
    /** A boolean that keeps track of if the grid is on or off. */
    private Boolean myGridIsOn;
    
    /** A boolean that keeps track of if the list is cleared or not. */
    private boolean myDraggedShapeIsShown;
    
    /**
     * This constructor initializes all of the variables and makes method calls to set size 
     * and background of panel as well as adding the mouse listeners.
     */
    public DrawingPanel() {
        
        super();
        myCurrentTool = new PencilTool(this);
        myBrushColor = Color.BLACK;
        myBrushWidth = 1;
        
        myShapes = new ArrayList<>();
        myUndoStack = new ArrayDeque<PaintShape>();
        myRedoStack = new ArrayDeque<PaintShape>();
        
        setBooleans();
        setSizeAndBackground();
        addMouseListeners();
    }
    
    /**
     * This method initializes the collections used in this class.
     */
    private void setBooleans() {
        myGridIsOn = false;
        myDraggedShapeIsShown = false;
    }
    
    /**
     * this method sets the size and initial background color.
     */
    private void setSizeAndBackground() {
        this.setPreferredSize(new Dimension(WIDTH, HEIGHT));
        this.setBackground(CANVAS_COLOR);
    }
    
    /**
     * This method adds the mouse listeners to the panel. 
     */
    private void addMouseListeners() {
        final MouseListeners allTheListeners = new MouseListeners();
        this.addMouseListener(allTheListeners);
        this.addMouseMotionListener(allTheListeners);
    }
    
    /**
     * This method is called whenever repaint() is used. It is used to redraw current or past 
     * images.
     * 
     * @param theGraphics the graphics used to draw objects.
     */
    @Override
    public void paintComponent(final Graphics theGraphics) {
        
        super.paintComponent(theGraphics);
        final Graphics2D graphics = (Graphics2D) theGraphics;
        
        graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                             RenderingHints.VALUE_ANTIALIAS_ON);

        drawPreviousShapes(graphics);

        if (myDraggedShapeIsShown) {
            drawCurrentShape(graphics);
        }
        if (myGridIsOn) {   
            paintGrid(graphics);
        } 
    }
    
    /**
     * This method draws all of the previous PaintShapes that have been made.
     * 
     * @param theGraphics the graphics used to draw.
     */
    private void drawPreviousShapes(final Graphics2D theGraphics) {
        
        for (final PaintShape s : myShapes) {
            
            theGraphics.setPaint(s.getColor());
            theGraphics.setStroke(s.getShapeWidth());
            theGraphics.draw(s.getShape());
        }
    }
    
    /**
     * this method draws the current PaintShape being drawn.
     * 
     * @param theGraphics the graphics used to draw.
     */
    private void drawCurrentShape(final Graphics2D theGraphics) {
        
        theGraphics.setPaint(myCurrentTool.returnDraggedShape().getColor());
        theGraphics.setStroke(myCurrentTool.returnDraggedShape().getShapeWidth());
        theGraphics.draw(myCurrentTool.returnDraggedShape().getShape());
    }
    
    /**
     * This method allows a grid to be drawn on the canvas.
     * 
     * @param theGraphics the graphics used to draw
     */
    private void paintGrid(final Graphics2D theGraphics) {
    
        theGraphics.setPaint(Color.GRAY);
        theGraphics.setStroke(new BasicStroke(1));
        
        for (int i = 0; i < getSize().width; i += GRID_WIDTH) {
            final Shape line = new Line2D.Float(i, 0, i, getSize().height);
            theGraphics.draw(line);
        }

        for (int i = 0; i < getSize().height; i += GRID_WIDTH) {
            final Shape line = new Line2D.Float(0, i, getSize().width, i);
            theGraphics.draw(line);
        }
    }

    /**
     * This method sets the current tool being used to draw.
     * 
     * @param theTool the current tool
     */
    public void setCurrentTool(final Tool theTool) {
        
        myCurrentTool = theTool;
    }
    
    /**
     * This method sets the current color to a new color that is choosed.
     * 
     * @param theBrushColor the new color
     */
    public void setBrushColor(final Color theBrushColor) {
        
        myBrushColor = theBrushColor;
    }
    
    /**
     * This method sets the new brush width that will be used to draw shapes.
     * 
     * @param theBrushWidth the new brush width
     */
    public void setBrushWidth(final int theBrushWidth) {
        
        myBrushWidth = theBrushWidth;
    }
    
    /**
     * This method returns the current color selection.
     * 
     * @return the current brush color
     */
    public Color getBrushColor() {
        
        return myBrushColor;
    }
    
    /**
     * This method returns the current brush width.
     * 
     * @return the current width of the drawing tool
     */
    public int getBrushWidth() {
        
        return myBrushWidth;
    }
    
    /**
     * This method clears the ArrayList used to stare previously drawn shapes. it is used to 
     * erase all objects on canvas.
     */
    public void clearLists() {
        
        myDraggedShapeIsShown = false;
        firePropertyChange(REDO_TOGGLE, true, false);
        firePropertyChange(UNDO_TOGGLE, true, false);
        myShapes.clear();
        myUndoStack.clear();
        myRedoStack.clear();
    }
    
    /**
     * This method flips the status of if the grid is on or off.
     */
    public void setGridStatus() {
        
        myGridIsOn ^= true;
    }
    
    /**
     * This method controls the behavior of when a shape is redrawn when the redo button is 
     * pressed.
     */
    public void undoDrawnShape() {
        
        if (!myShapes.isEmpty()) {
            
            myShapes.remove(myShapes.size() - 1);
        } 
        if (myShapes.isEmpty()) {
            
            firePropertyChange(UNDO_TOGGLE, true, false);
        } 
        if (!myUndoStack.isEmpty()) {
            
            myRedoStack.push(myUndoStack.pop());
        }
        
        firePropertyChange(REDO_TOGGLE, false, true);
    }
    
    /**
     * This method controls the behavior of when a shape is redrawn when the undo button is 
     * pressed.
     */
    public void redoDrawnShape() {
        
        PaintShape tempShape = null;
        
        if (!myRedoStack.isEmpty()) {
            
            tempShape = myRedoStack.pop();
            myShapes.add(tempShape);
            myUndoStack.push(tempShape);
        }
        
        if (myRedoStack.isEmpty()) {
            
            firePropertyChange(REDO_TOGGLE, true, false);
        }
        
        firePropertyChange(UNDO_TOGGLE, false, true);
    }
    
    /**
     * This class allows a MouseListener to be instantiated and used on the JPanel. It also 
     * makes calls to the current tool to implement the behavior of that tool when mouse is 
     * pressed dragged or released.
     */
    private class MouseListeners extends MouseInputAdapter {
        
        @Override
        public void mousePressed(final MouseEvent theAction) {
            
            myCurrentTool.ifPressed(theAction);
        }
        
        @Override
        public void mouseDragged(final MouseEvent theAction) {
            
            myDraggedShapeIsShown = true;
            myCurrentTool.whileDragged(theAction);
            repaint();
        }
        
        @Override
        public void mouseReleased(final MouseEvent theAction) {
            
            myCurrentTool.ifReleased(theAction);
            myShapes.add(myCurrentTool.returnReleasedShape());
            myDraggedShapeIsShown = false;
            myUndoStack.push(myCurrentTool.returnReleasedShape());
            firePropertyChange(UNDO_TOGGLE, false, true);
            repaint();
        }  
    }
}